/**********************************************************************
 * File: bsp.cpp
 * Version: 0.1
 * Project: 
 * Created: 12 Jan
 * Last modified:
 *
 * Comment: 
 * void setIntMask(int mask);    : set interrupt to mask
 * unsigned int enterCS(void);   : switch to SV and disable 
 *                                 both fast and normal int,
 *                                 return old state of mode 
 *                                 and int
 * unsigned int enterCS(void);   : restore old state
 *
 * Author: Jens-Christian Lache
 * Fachbereich: TI3 Technische Informatik
 * Prof. Dr. Thomas Teufel
 * http://www.ti3.tu-harburg.de/deutsch/
 **********************************************************************/



#include "bsp.h"
#include <stdio.h>

/**********************************************************************
 * Function: setIntMask
 * Object: set the interrupt DISABLE bits 
 * in the "current program status register" (cpsr) according to "mask"
 * created: 17 Jan
 * last modified: 
 **********************************************************************/

void setIntMask(int mask) {
  int volatile previousMode;
  int volatile cpsr;

#ifdef TARGET_SIM
  return;
#endif // TARGET_SIM

#ifdef NO_SWI
  return;
#endif // NO_SWI

  // some .equ's for better understanding
  asm volatile (".equ SR_I_FIELD,       0xc0 @ both interrupts");
  asm volatile (".equ SR_F_BIT,         0x40 @ Fast interrupt bit");
  asm volatile (".equ SR_I_BIT,         0x80 @ Interrupt Request bit");
  asm volatile (".equ EXEC_CRIT_SR,     (SR_F_BIT|SR_I_BIT) @ critical section int mask");
  asm volatile (".equ MODE_BITS,        0x1F @ all mode bits from cpsr");
  asm volatile (".equ angel_SWIreason_EnterSVC, 0x17");
  asm volatile (".equ angel_SWI_ARM, 0x123456");

  // get mode
  asm volatile ("mrs %0, CPSR":"=r" (cpsr));
  previousMode = (cpsr & MODE_BITS);

  // switch to SV
  asm volatile ("mov r0, #angel_SWIreason_EnterSVC; swi #angel_SWI_ARM": : 
		: "r0", "r1", "r2", "r3", "r4", "r5","r6","r7","r8","r9","r10","ip", "lr", "memory");
  // Clobbers r0 and r1, and lr if in supervisor mode 
  // Accordingly to page 13-77 of ARM DUI 0040D other registers
  // can also be clobbered.  Some memory positions may also be
  // changed by a system call, so they should not be kept in
  // registers. Note: we are assuming the manual is right and
  // Angel is respecting the APCS

  // we are now in SV, ints are disabled
  asm volatile ("mrs     %0, CPSR":"=r" (cpsr));

  // mask out the two int bits
  asm volatile ("bic     %0, %1, #SR_I_FIELD":"=r" (cpsr):"r" (cpsr));

  // mask out the five mode bits
  asm volatile ("bic     %0, %1, #MODE_BITS":"=r" (cpsr):"r" (cpsr));

  // set the mask and switch to the previous mode
  cpsr = (cpsr | mask | previousMode);

  // write back to CPSR
  // asm volatile("msr CPSR, %0": :"r" (cpsr));
}

/**********************************************************************
 * Function: setModeMask
 * Object: set the mode bits in the 
 * "current program status register" (cpsr) according to "mask"
 * created: 17 Jan
 * last modified: 
 **********************************************************************/

void setModeMask(int mask) {
  int volatile previousInt;
  int volatile cpsr;

#ifdef TARGET_SIM
  return;
#endif

#ifdef NO_SWI
  return;
#endif // NO_SWI



  asm volatile (".equ SR_I_FIELD,       0xc0 @ both interrupts");
  asm volatile (".equ SR_F_BIT,         0x40 @ Fast interrupt bit");
  asm volatile (".equ SR_I_BIT,         0x80 @ Interrupt Request bit");
  asm volatile (".equ EXEC_CRIT_SR,     (SR_F_BIT|SR_I_BIT) @ critical section int mask");
  asm volatile (".equ MODE_BITS,        0x1F @ all mode bits from cpsr");
  asm volatile (".equ angel_SWIreason_EnterSVC, 0x17");
  asm volatile (".equ angel_SWI_ARM, 0x123456");

  // get int state
  asm volatile ("mrs %0, CPSR":"=r" (cpsr));
  previousInt = (cpsr & SR_I_FIELD);

  // switch to SV
  asm volatile ("mov r0, #angel_SWIreason_EnterSVC; swi #angel_SWI_ARM": : 
		: "r0", "r1", "r2", "r3", "r4", "r5","r6","r7","r8","r9","r10","ip", "lr", "memory");
  // Clobbers r0 and r1, and lr if in supervisor mode 
  // Accordingly to page 13-77 of ARM DUI 0040D other registers
  // can also be clobbered.  Some memory positions may also be
  // changed by a system call, so they should not be kept in
  // registers. Note: we are assuming the manual is right and
  // Angel is respecting the APCS

  // we are now in SV, ints are disabled
  asm volatile ("mrs     %0, CPSR":"=r" (cpsr));

  // mask out the two int bits
  asm volatile ("bic     %0, %1, #SR_I_FIELD":"=r" (cpsr):"r" (cpsr));

  // mask out the five mode bits
  asm volatile ("bic     %0, %1, #MODE_BITS":"=r" (cpsr):"r" (cpsr));

  cpsr = (cpsr | mask | previousInt);

  // asm volatile ("msr CPSR, %0": :"r" (cpsr));
}

/**********************************************************************
 * Function: enterCS 
 * Object: disable interrupts and switch mode to supervisor,
 * returns a mask containing the previos interrupt and mode state
 * created: 12 Jan
 * last modified: 
 **********************************************************************/

unsigned int enterCS(void){
  int volatile intState;
  int volatile modeState;
  int volatile returnValue;

#ifdef TARGET_SIM
  return 1;
#endif

#ifdef NO_SWI
  return 1;
#endif //NO_SWI


  asm volatile (".equ SR_I_FIELD,       0xc0 @ both interrupts");
  asm volatile (".equ SR_F_BIT,         0x40 @ Fast interrupt bit");
  asm volatile (".equ SR_I_BIT,         0x80 @ Interrupt Request bit");
  asm volatile (".equ EXEC_CRIT_SR,     (SR_F_BIT|SR_I_BIT) @ critical section int mask");
  asm volatile (".equ angel_SWIreason_EnterSVC, 0x17");
  asm volatile (".equ angel_SWI_ARM, 0x123456");

  asm volatile ("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@");
  asm volatile ("@ This code should turn off interrupts to prevent angel from running");
  asm volatile ("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@");

  // get the old state of I and F flag
  // and save it in "intState" respectivly
  // get old state of mode and save it in
  // "modeState"
  asm volatile("mrs r0, CPSR @ get CPSR": : :"r0");
  asm volatile("and %0,r0, #SR_I_FIELD":"=r" (intState));
  asm volatile("and %0,r0, #MODE_BITS":"=r" (modeState));
  returnValue = (intState|modeState);
  // use angel to change to SV
  // angel will also disable interrupts
  asm volatile ("mov r0, #angel_SWIreason_EnterSVC; swi #angel_SWI_ARM": : 
		 : "r0", "r1", "r2", "r3", "r4", "r5", "r6","r7","r8","r9","r10","ip", "lr", "memory");
  // Clobbers r0 and r1, and lr if in supervisor mode 
  // Accordingly to page 13-77 of ARM DUI 0040D other registers
  // can also be clobbered.  Some memory positions may also be
  // changed by a system call, so they should not be kept in
  // registers. Note: we are assuming the manual is right and
  // Angel is respecting the APCS

  // return previous interrupt state / mask
  return returnValue;
}
/**********************************************************************
 * Function: exitCS
 * Object: restores the interrupt and mode state previous of
 * "enterCS". This function assumes, that "enterCS" has been called
 * and that we are in superuser mode.
 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 * !!CAUTION: This function will crash, if we are not in superuser mode!!
 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 * created: 12 Jan
 * last modified: 
 **********************************************************************/

void exitCS(unsigned int restoreValue){
#ifdef TARGET_SIM
  return;
#endif

#ifdef NO_SWI
  return;
#endif //NO_SWI



  asm volatile (".equ SR_I_FIELD,       0xc0 @ both interrupts");
  asm volatile (".equ SR_F_BIT,         0x40 @ Fast interrupt bit");
  asm volatile (".equ SR_I_BIT,         0x80 @ Interrupt Request bit");
  asm volatile (".equ MODE_BITS,        0x1F @ all mode bits from cpsr");
  asm volatile (".equ USER_MODE,        0x10 @ b10000 is user mode");
  asm volatile (".equ EXEC_CRIT_SR,     (SR_F_BIT|SR_I_BIT) @ critical section int mask");

    // switch back to previous mode and int state
  asm volatile ("mrs r0, CPSR": : :"r0");
  asm volatile ("bic r0, r0, #MODE_BITS": : :"r0");
  asm volatile ("bic r0, r0, #SR_I_FIELD": : :"r0");
  asm volatile ("orr r0, r0, %0": :"r" (restoreValue));
  // asm volatile ("msr CPSR, r0");
}





// int value;
//   asm volatile ("mov r0, %1; mov r1, %2; swi %a3; mov %0, r0"
//        : "=r" (value) /* Outputs */
//        : "r" (reason), "r" (arg), "i" (AngelSWI) /* Inputs */
//        : "r0", "r1", "r2", "r3", "ip", "lr", "memory", "cc"
// 		/* Clobbers r0 and r1, and lr if in supervisor mode */);
//                 /* Accordingly to page 13-77 of ARM DUI 0040D other registers
//                    can also be clobbered.  Some memory positions may also be
//                    changed by a system call, so they should not be kept in
//                    registers. Note: we are assuming the manual is right and
//                    Angel is respecting the APCS.  */
//   return value;














// unsigned int enterCS(int * p_){
//   int jumpToUserMode;
//   asm volatile (".equ SR_I_FIELD,       0xc0 @ both interrupts");
//   asm volatile (".equ SR_F_BIT,         0x40 @ Fast interrupt bit");
//   asm volatile (".equ SR_I_BIT,         0x80 @ Interrupt Request bit");
//   asm volatile (".equ EXEC_CRIT_SR,     (SR_F_BIT|SR_I_BIT) @ critical section int mask");
//   asm volatile (".equ angel_SWIreason_EnterSVC, 0x17");
//   asm volatile (".equ angel_SWI_ARM, 0x123456");

//   /***************************************************
//    * Code from Linux kernel 2.4
//    * mov     r0, #0x17               @ angel_SWIreason_EnterSVC
//    * swi     0x123456                @ angel_SWI_ARM
//    * mrs     r0, cpsr                @ turn off interrupts to
//    * orr     r0, r0, #0xc0           @ prevent angel from running
//    * msr     cpsr_c, r0
//    **************************************************/

//   /***************************************************
//    * Code from Quentin Barnes
//    * Here's my disable function:
//    * ====
//    * DEF_ARM_FN   disable_int, DUAL_STATE
//    * mrs     r0, CPSR
//    * orr     r1, r0, #EXEC_CRIT_SR
//    * msr     CPSR, r1
//    * bx      lr
//    * ====
//    * 
//    * It has the C prototype: "unsigned int disable_int(void)".
//    * 
//    * The restore function:
//    * ====
//    * DEF_ARM_FN  restore_int, DUAL_STATE
//    * 
//    *  and     r1, r0, #SR_I_FIELD
//    * mrs     r0, CPSR
//    *  bic     r2, r0, #SR_I_FIELD
//    *  orr     r1, r1, r2
//    * msr     CPSR, r1
//    * bx      lr
//    * ====
//    * 
//    * It has the C prototype: "unsigned int restore_int(unsigned int)".
//   **************************************************/
// #ifdef SWI_OK

//   asm volatile ("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@");
//   asm volatile ("@ This code should turn off interrupts to prvent angel from running");
//   asm volatile ("@ All instructions containing mrs and msr will crash, when not in SV mode!!");
//   asm volatile ("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@");

//   //asm volatile ("mrs %0, CPSR    @ get old PSR (Program Status Reg)":"=r" (storeCPSR));
//   //asm volatile ("mov r0, #angel_SWIreason_EnterSVC": : :"r0");

//   asm volatile ("swi #angel_SWI_ARM": : :"r0");
//   asm volatile ("mov %0,r0 @ the call to EnterSVC returns the address of ReturnToUser":"=r" (tmpVar));
//   * p_jumpToUserMode = tmpVar;
//   // the int should already be dissabled (by the previos swi) so the next lines are useless
//   asm volatile ("mrs r0, CPSR    @ get PSR ": : :"r0" );
//   asm volatile ("orr r1, r0, #EXEC_CRIT_SR @ mask in the two int disable bits": : :"r1");
//   asm volatile ("msr CPSR,r1 @write back to PSR"); 
  
//   // return old value of CPSR
//   return storeCPSR;
// #endif // SWI_OK
  
// }
// unsigned int exitCS(unsigned int restoreValue){
//   asm volatile (".equ SR_I_FIELD,       0xc0 @ both interrupts");
//   asm volatile (".equ SR_F_BIT,         0x40 @ Fast interrupt bit");
//   asm volatile (".equ SR_I_BIT,         0x80 @ Interrupt Request bit");
//   asm volatile (".equ EXEC_CRIT_SR,     (SR_F_BIT|SR_I_BIT) @ critical section int mask");

// #ifdef SWI_OK
//   asm volatile ("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@");
//   asm volatile ("@ All instructions containing mrs and msr will crash, when not in SV mode!!");
//   asm volatile ("@ r0 contains restoreValue");
//   asm volatile ("@ r0 will be returned");
//   asm volatile ("@ do I want to switch back to USR mode here?");
//   asm volatile ("@ if so, how do I call ReturnToUSR?");
//   asm volatile ("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@");

//   asm volatile ("and r1,%0, #SR_I_FIELD @ mask out disable bit if set in restoreValue": :"r" (restoreValue) :"r1");
//   asm volatile ("mrs %0,CPSR @ move restoreValue to PSR": :"r" (restoreValue));
//   asm volatile ("bic r2,%0, #SR_I_FIELD @ clear the ": :"r" (restoreValue) :"r2");
//   asm volatile ("orr r1,r1,r2": : :"r1");
//   asm volatile ("msr CPSR, r1 @ ");

//   // switch back to user mode
//   asm volatile ("mov r1,%0": :"r" (tmpVar):"r1");
//   asm volatile ("mov lr,pc");
//   asm volatile ("mov pc,r1");

// #endif // SWI_OK 
// }

// void setIntMask(int mask) {
//   int previousMode;
//   int r0;
//   int cpsr;
//   asm volatile (".equ SR_I_FIELD,       0xc0 @ both interrupts");
//   asm volatile (".equ SR_F_BIT,         0x40 @ Fast interrupt bit");
//   asm volatile (".equ SR_I_BIT,         0x80 @ Interrupt Request bit");
//   asm volatile (".equ EXEC_CRIT_SR,     (SR_F_BIT|SR_I_BIT) @ critical section int mask");
//   asm volatile (".equ MODE_BITS,        0x1F @ all mode bits from cpsr");
//   asm volatile (".equ angel_SWIreason_EnterSVC, 0x17");
//   asm volatile (".equ angel_SWI_ARM, 0x123456");

//   // printf("\n******setIntMask(mask = %X): *****\n",mask);
  

//   // debug
//   // asm volatile ("mrs %0, CPSR":"=r" (cpsr));
//   // printf("setIntMask: originally cpsr at beginning = %X\n",cpsr);
//   // end debug
//   // get mode
//   asm volatile ("mrs r0, CPSR": : :"r0");
//   asm volatile ("and %0, r0, #MODE_BITS":"=r" (previousMode));
//   // switch to SV
//   asm volatile ("mov r0, #angel_SWIreason_EnterSVC": : :"r0");
//   asm volatile ("swi #angel_SWI_ARM": : :"r0");
//   // we are now in SV, ints are disabled
//   asm volatile ("mrs     r0, CPSR": : :"r0");
//   // mask out the two int bits
//   asm volatile ("bic     r0, r0, #SR_I_FIELD": : :"r0");
//   // mask out the five mode bits
//   asm volatile ("bic     r0, r0, #MODE_BITS": : :"r0");

//   // // debug
// //   asm volatile ("mov %0, r0": "=r" (r0));
// //   // printf("setIntMask: r0 = %X\n",r0);
// //   asm volatile ("mrs     r0, CPSR": : :"r0");
// //   // mask out the two int bits
// //   asm volatile ("bic     r0, r0, #SR_I_FIELD": : :"r0");
// //   // mask out the five mode bits
// //   asm volatile ("bic     r0, r0, #MODE_BITS": : :"r0");
// //   // debug end

//   // set the mask
//   asm volatile ("orr     r0, r0, %0": :"r" (mask));
//   // switch to the previous mode
//   asm volatile ("orr     r0, r0, %0": :"r" (previousMode));

//   // debug
//   // asm volatile ("mrs %0, CPSR":"=r" (cpsr));
//   asm volatile ("mov %0, r0": "=r" (r0));
//   // printf("setIntMask: mask = %X\n",mask);
//   // printf("setIntMask; previousMode = %X\n",previousMode);
//   // printf("setIntMask: cpsr = %X\n",cpsr);
//   // printf("setIntMask: r0 = %X\n",r0);
//   // debug end
//   asm volatile("msr CPSR, %0": :"r" (r0));

//   //asm volatile ("msr     CPSR, r0");
//   // printf("******setIntMask End *****\n\n\n");
