/**********************************************************************
 *
 * Filename:    task.cpp
 * 
 * Description: An implementation of real-time tasks for ADEOS.
 *
 * Notes:       
 *
 * 
 * Copyright (c) 1998 by Michael Barr.  This software is placed into
 * the public domain and may be used for any purpose.  However, this
 * notice must not be changed or removed and no warranty is either
 * expressed or implied by its publication or distribution.
 **********************************************************************/
#include "stdio.h"
#include "adeos.h"
#include "memAllocator.h"
extern void contextInit(Context *, void (*run)(Task *), Task *, int * pStackTop);
extern void contextSwitch(Context * pOldContext, Context * pNewContext);
extern MemAllocator internalMem;

TaskId    Task::nextId = 0;


/**********************************************************************
 *
 * Method:      TaskList()
 *
 * Description: Create a new linked list of tasks.
 *
 * Notes:	    
 * 
 * Returns:	    None defined.
 *
 **********************************************************************/
TaskList::TaskList()
{
    pTop = NULL;

}   /* TaskList() */


/**********************************************************************
 *
 * Method:      insert()
 *
 * Description: Insert a task into an ordered linked list.
 *
 * Notes:	    The caller is responsible for disabling interrupts.
 * 
 * Returns:	    None defined.
 *
 **********************************************************************/
void TaskList::insert(Task * pTask) 
{ 
  Task ** ppPrev = &(this->pTop); 

  
  // 
  // Handle  the case of an empty list. 
  // 
  if (*ppPrev == NULL) // if this->pTop == NULL
    { 
      *ppPrev = pTask; // then this->pTop = pTask;
      return;  
    } 
  
  // 
  // Walk down the ordered list until a lower priority task is found. 
  // 
  
  
  while ((*ppPrev)->pNext != NULL &&
   	 pTask->priority <= (*ppPrev)->pNext->priority) { 
    ppPrev  = &(*ppPrev)->pNext; 
  } 

  // Handle the case that we have to insert before the first element
  if (ppPrev == &(this->pTop) && (pTask->priority > (*ppPrev)->priority)) { 
                                            // new task has higher priority
      pTask->pNext = (*ppPrev);             // insert between pTop and 
      this->pTop = pTask;                   // first element of the list
      return;
  }
  // 
  // Insert the new task at this point in the linked list.  
  // 
  pTask->pNext = (*ppPrev)->pNext; 
  (*ppPrev)->pNext = pTask;

  
        
} /* insert() */ 



/**********************************************************************
 *
 * Method:      remove()
 *
 * Description: Remove a task from a linked list.
 *
 * Notes:	    The caller is responsible for disabling interrupts.
 * 
 * Returns:	    A pointer to the removed task.  Or NULL if it wasn't
 *              found in the linked list.
 *
 **********************************************************************/
Task *
TaskList::remove(Task * pTask)
{
    Task **  ppPrev = &this->pTop;


    //
    // Walk down the linked list until the dead task is found.
    //
    while (*ppPrev != NULL && *ppPrev != pTask)
    {
        ppPrev = &(*ppPrev)->pNext;
    }

    //
    // Ensure that the task was actually found.
    //
    if (*ppPrev == NULL)
    {
        return (NULL);
    }

    //
    // Remove the task from the linked list.
    //
    *ppPrev = pTask->pNext;

    return (pTask);

}   /* remove() */


/**********************************************************************
 *
 * Function:    run()
 *
 * Description: Start a task and cleanup after it if it exits.
 *
 * Notes:       
 * 
 * Returns:     None defined.
 *
 **********************************************************************/
void
run(Task * pTask)
{
  int intAndModeState;
    //
    // Start the task, by executing the associated function.
    //
    pTask->entryPoint();

    intAndModeState=enterCS();                          ////// Critical Section Begin

    // jc
    // ???? Where is the corresponding exitCS() function to this line???

    //
    // Remove this task from the scheduler's data structures.
    //

    os.readyList.remove(pTask);
    os.pRunningTask = NULL;

    //
    // Free the task's stack space.
    //
#ifdef AT91M40400
    delete pTask->pStack;
#endif // AT91M40400
#ifdef TARGET_SIM
    if (internalMem.deleteMem(pTask->pStack) == FALSE) {
      // Error
      return;
    }
#endif // TARGET_SIM

    os.schedule();                          // Scheduling Point

    // This line will never be reached.

}   /* run() */


/**********************************************************************
 *
 * Method:      Task()
 *
 * Description: Create a new task and initialize its state.
 *
 * Notes:
 * 
 * Returns:     
 *
 **********************************************************************/
Task::Task(void (*function)(), Priority p, int stackSize) {
  int intAndModeState;
  stackSize /= sizeof(int);               // Convert bytes to words.
  
  intAndModeState = enterCS();                          ////// Critical Section Begin
  
  // 
  // Initialize the task-specific data.
  //
  id         = Task::nextId++;
  state      = Ready;
  priority   = p;        
  entryPoint = function;
#ifdef AT91M40400
  pStack     = new int[stackSize];
#endif // AT91M40400
  // The simulator does not know about the monitor program
  // "angel". "new int[stackSize];" does somewhere make a
  // call to the swi handler of angel. memAllocator will
  // get a decent amount of mem and allocate parts of it
  // to the callers of "memAllocator.new(int );"
#ifdef TARGET_SIM
  pStack     = internalMem.newMem(stackSize);
#endif // TARGET_SIM
  pNext      = NULL;

  // debug
#ifdef WITH_PRINTF
  printf("&context = 0x%X\n", (int) &context );
  printf("run = 0x%X\n", (int) run );
  printf("this = 0x%X\n", (int) this );
  printf("pStack + stackSize = 0x%X\n", (int)  (pStack + stackSize));
#endif // WITH_PRINTF
  // debug ende

  //
  // Initialize the processor context.
  //
  contextInit(&context, run, this, pStack + stackSize);    

  //
  // Insert the task into the ready list.
  //
  os.readyList.insert(this);
  
  os.schedule();                          // Scheduling Point
  
  exitCS(intAndModeState);                           ////// Critical Section End

}   /* Task() */
